package com.wawrzek.fumbo;

import jade.content.lang.*;
import jade.content.lang.sl.*;
import jade.content.onto.*;
import jade.core.*;

import java.io.*;

import org.apache.log4j.*;

import com.wawrzek.fumbo.coordinator.*;
import com.wawrzek.fumbo.core.*;
import com.wawrzek.fumbo.gui.*;
import com.wawrzek.fumbo.tools.*;
import com.wawrzek.fumbo.tools.basic.*;

/**
 * Main class of coordinator Agent. Coordinator takes care about workers (and their jobs) and communicates 
 * with rest of the world
 * 
 * @author Wawrzek
 */
public class Coordinator extends Agent
{
    /** Coordinator's GUI */
    private CoordinatorGui myGui;
    /** Coordinator's configuration */
    private Record config;

    /** Used codec */
    private Codec codec = new SLCodec();
    /** Ontolgy used to communicate with workers */
    private Ontology ontology = InternalOntology.getInstance();
    /** Ontolgy used to communicate with outside agents */
    private Ontology externalOntology = ExternalOntology.getInstance();

    /** manager that takes care of tasks to do */
    private TaskManager taskManager;
    /** manager of worker agents */
    private AgentsManager agentsManager;
    /** logger where all info is logged */
    private Logger logger;

    
    /**
     * Setups cootdinator - its behabiours, logger etc.
     */
    protected void setup()
    {
        try
        {
            setupOut();
            startConfig();
            setupCore();
            myGui = new CoordinatorGui(this);
            setBehaviours();
            myGui.show();
            logger.info("\n\n---------------------------------------------");
            logger.info(getAID().getName() + " is ready.");
            myGui.getLogTextPane().refresh();
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            logger.fatal("Abnormal agent termination: " + ex.getMessage());
            doDelete();
        }
    }

    /**
     * Prepares standard output and logging
     * 
     * @throws FileNotFoundException
     */
    private void setupOut() throws FileNotFoundException
    {
        PropertyConfigurator.configure("cfg/logging.cfg");
        logger = Logger.getLogger(Consts.COORDINATOR_LOG);

        PrintStream out = new PrintStream(new FileOutputStream("out.txt"));
        System.setOut(out);
        System.setErr(out);
    }

    /**
     * Reads config from a file
     * 
     * @throws AgentException
     */
    private void startConfig() throws AgentException
    {
        config = new Record();
        if (config.read("cfg" + File.separator + "fumbo.cfg") == false)
            throw new AgentException("error while reading config file");
        Logger.getLogger(Consts.COORDINATOR_LOG).debug("congig: " + config.toString());
    }

    /**
     * Prepares ontologies and managers
     *
     */
    private void setupCore()
    {
        getContentManager().registerLanguage(codec);
        getContentManager().registerOntology(ontology);
        getContentManager().registerOntology(externalOntology);
        getContentManager().setValidationMode(true);
        taskManager = new TaskManager(codec, ontology, getAID());
        agentsManager = new AgentsManager(this, config.getString(Consts.FLD_RULEDIRECTORY));
    }

    /**
     * Starts all cyclic behaviours
     *
     */
    private void setBehaviours()
    {
        addBehaviour(new TaskRequestLoop());
        addBehaviour(new AnswerCollectLoop());
        addBehaviour(new LookForWorkersLoop(this, config.getInt(Consts.FLD_WORKERSPERIOD)));
        addBehaviour(new WorkersFailureLoop());
        if (config.getInt(Consts.FLD_WATCHLOG) == 1)
            addBehaviour(new RefreshLogLoop(this, myGui.getLogTextPane(), 1000));
        
    }

    protected void takeDown()
    {
        agentsManager.killAllWorkers();
        logger.info("Bye bye!");
        LogManager.shutdown();
    }

    public AgentsManager getAgentsManager()
    {
        return agentsManager;
    }

    public TaskManager getTaskManager()
    {
        return taskManager;
    }

    public Record getConfig()
    {
        return config;
    }

    public void setConfig(Record config)
    {
        this.config = config;
    }

    public Ontology getOntology()
    {
        return ontology;
    }

    public Ontology getExternalOntology()
    {
        return externalOntology;
    }

    public Codec getCodec()
    {
        return codec;
    }

    public CoordinatorGui getGui()
    {
        return myGui;
    }
}