package com.wawrzek.fumbo.coordinator;

import jade.core.*;
import jade.util.leap.*;
import jade.wrapper.*;
import jade.wrapper.AgentContainer;

import java.io.*;

import org.apache.log4j.*;

import com.wawrzek.fumbo.*;
import com.wawrzek.fumbo.coordinator.*;
import com.wawrzek.fumbo.gui.*;
import com.wawrzek.fumbo.tools.*;

/**
 * @author wawrzek
 * 
 * Manager that keeps all information about working TuningAgents
 */
public class AgentsManager
{
    /** Coordinator - owner of manager */
    private Coordinator coordinator;
    /** list of workers */
    private ArrayList agents;
    /** name of directory where rules files are */
    private String directory;

    /** model for GUI JTable */
    private WorkersTableModel tableModel;

    
    public AgentsManager(Coordinator coordinator, String directory)
    {
        this.coordinator = coordinator;
        agents = new ArrayList();
        this.directory = directory;
    }

    /**
     * Creates new agents according to rule files in rule directory
     * @throws AgentException if something went bad
     */
    public void makeNewAgents() throws AgentException
    {
        try
        {
            File dir = new File(directory);
            if (!dir.exists())
                throw new AgentException("no " + directory + " directory");
            if (!dir.isDirectory())
                throw new AgentException(directory + " is not a directory");

            File[] files = dir.listFiles();

            for (int i = 0; i < files.length; i++)
            {
                File file = files[i];
                if (!agentExists(file.getName()))
                    addAgent(file.getName());
            }
        }
        catch (Exception ex)
        {
            Logger.getLogger(Consts.COORDINATOR_LOG).error("error while making new workers", ex);
            ex.printStackTrace();
        }
    }

    /**
     * Checks if agent exists (on Coordinator's list, not if he lives in Jade)
     * @param wi info about worker
     * @return does agent exist
     */
    private boolean agentExists(WorkerInfo wi)
    {
        File dir = new File(directory);
        File[] files = dir.listFiles();

        for (int i = 0; i < files.length; i++)
        {
            if (wi.getFileName().equals(files[i].getName()))
                return true;
        }
        return false;
    }

    /**
     * Checks if agent exists (on Coordinator's list, not if he lives in Jade)
     * @param fileName rule-filename of the agent
     * @return
     */
    private boolean agentExists(String fileName)
    {
        Iterator itr = agents.iterator();
        while (itr.hasNext())
        {
            WorkerInfo wi = (WorkerInfo) (itr.next());
            if (fileName.equals(wi.getFileName()))
                return true;
        }
        return false;
    }

    /**
     * Creates an agent according to rule file
     * @param fileName name of a rule-file (in the rule-directory)
     * @throws StaleProxyException in case of error while creating a new agent
     */
    private void addAgent(String fileName) throws StaleProxyException
    {
        AgentContainer ac = coordinator.getContainerController();
        Object args[] = new Object[2];
        args[0] = directory + File.separator + fileName;
        args[1] = coordinator.getAID();

        int dotindex = fileName.indexOf(".");
        String agentName = fileName;
        if (dotindex != -1)
            agentName = fileName.substring(0, dotindex);
        agentName = "[" + agentName + "]";

        AgentController newAgent = ac.createNewAgent(agentName, "com.wawrzek.fumbo.Worker", args);
        newAgent.start();

        WorkerInfo wi = new WorkerInfo(fileName, newAgent.getName());
        agents.add(wi);
        tableModel.fireTableDataChanged();
    }

    /**
     * kills all agents that are on coordinator's list. 
     */
    public void killAllWorkers()
    {
        Iterator itr = agents.iterator();
        try
        {
            AgentContainer ac = coordinator.getContainerController();
            while (itr.hasNext())
            {
                WorkerInfo wi = (WorkerInfo) (itr.next());
                Logger.getLogger(Consts.COORDINATOR_LOG).info("killing agent " + wi.getAid());
                tableModel.fireTableDataChanged();
                ac.getAgent(wi.getName()).kill();
                itr.remove();
            }
        }
        catch (StaleProxyException e)
        {
            Logger.getLogger(Consts.COORDINATOR_LOG).error("error while killing agents: " + e.getMessage());
            e.printStackTrace();
        }
        catch (ControllerException e)
        {
            Logger.getLogger(Consts.COORDINATOR_LOG).error("error while killing agents: " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Kills all agents, that does not have their file now. For example, when somebody removed a rule-file 
     */
    public void killAbsentWorkers()
    {
        Iterator itr = agents.iterator();
        try
        {
            AgentContainer ac = coordinator.getContainerController();
            while (itr.hasNext())
            {
                WorkerInfo wi = (WorkerInfo) (itr.next());

                if (!agentExists(wi))
                {
                    Logger.getLogger(Consts.COORDINATOR_LOG).info("killing agent " + wi.getAid());
                    tableModel.fireTableDataChanged();
                    ac.getAgent(wi.getName()).kill();
                    itr.remove();
                }
            }
        }
        catch (StaleProxyException e)
        {
            Logger.getLogger(Consts.COORDINATOR_LOG).error("error while killing agents: " + e.getMessage());
            e.printStackTrace();
        }
        catch (ControllerException e)
        {
            Logger.getLogger(Consts.COORDINATOR_LOG).error("error while killing agents: " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Kills an agent of given aid
     * @param aid Id of agent to kill
     */
    public void killWorker(AID aid)
    {
        Iterator itr = agents.iterator();
        try
        {
            AgentContainer ac = coordinator.getContainerController();
            while (itr.hasNext())
            {
                WorkerInfo wi = (WorkerInfo) (itr.next());

                if (wi.getAid().equals(aid))
                {
                    Logger.getLogger(Consts.COORDINATOR_LOG).info("killing agent " + wi.getAid());
                    tableModel.fireTableDataChanged();
                    ac.getAgent(wi.getName()).kill();
                    itr.remove();
                    break;
                }
            }
        }
        catch (StaleProxyException e)
        {
            Logger.getLogger(Consts.COORDINATOR_LOG).error("error while killing agent: " + e.getMessage());
            e.printStackTrace();
        }
        catch (ControllerException e)
        {
            Logger.getLogger(Consts.COORDINATOR_LOG).error("error while killing agent: " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * That pings all agents on list. In fact it sends 2 signals: suspend and activate to everybody 
     * @return how many agents are there
     */
    public int pingAllAgents()
    {
        Iterator itr = agents.iterator();
        int result = 0;
        try
        {
            AgentContainer ac = coordinator.getContainerController();
            while (itr.hasNext())
            {
                WorkerInfo wi = (WorkerInfo) (itr.next());
                ac.getAgent(wi.getName()).suspend();
                ac.getAgent(wi.getName()).activate();
                result++;
            }
        }
        catch (StaleProxyException e)
        {
            e.printStackTrace();
            return -1;
        }
        catch (ControllerException e)
        {
            e.printStackTrace();
            return -1;
        }
        return result;
    }

    /**
     * returns a list of agents
     * @return a list of agents
     */
    public ArrayList getList()
    {
        return agents;
    }

    public WorkersTableModel getWorkersModel()
    {
        return tableModel;
    }

    public void setTableModel(WorkersTableModel wtm)
    {
        tableModel = wtm;
    }

}