package com.wawrzek.fumbo.coordinator;

import jade.content.lang.*;
import jade.content.lang.Codec.*;
import jade.content.onto.*;
import jade.content.onto.basic.*;
import jade.core.*;
import jade.lang.acl.*;
import jade.util.leap.*;

import org.apache.log4j.*;

import com.wawrzek.fumbo.*;
import com.wawrzek.fumbo.core.*;
import com.wawrzek.fumbo.tools.*;

/**
 * Class providing internal tasks management
 * @author Wawrzek
 *
 */
public class TaskManager
{
    /** Map containg list of current tasks */
    private HashMap tasks;
    /** Next number for next task (sequence) */
    private long task_seq;
    /** Codec used to write task */  
    private Codec codec;
    /** Ontlogy used to write task */
    private Ontology ontology;
    /** AID of coordinator to which final answer will be sent to */
    private AID coordinatorAID;

    /**
     * The only constructor
     * 
     * @param codec codec that will be used to read / write messages
     * @param ontology ontology that will be used to read / write messages
     * @param coordinatorAID AID of coordinator 
     */
    public TaskManager(Codec codec, Ontology ontology, AID coordinatorAID )
    {
        this.codec = codec;
        this.ontology = ontology;
        this.coordinatorAID = coordinatorAID;
        task_seq = 1;
        tasks = new HashMap();
    }

    /**
     * Registeres new task that came from outside the Fumbo.
     * 
     * @param etask external task that needs to be registered
     * @param agentList list of worker agents that will be asked to work
     * @param msg Request message that came with a new task 
     * @return number of registerd task
     * @throws AgentException in case of unresolvable problems
     */
    public long registerTask(ExternalTask etask, List agentList, ACLMessage msg) throws AgentException
    {
        Task task = new Task(etask, task_seq++, msg);
        task.setAgentsRing(agentList);
        task.setCoordinator(coordinatorAID);
        tasks.put(new Long(task.getId()), task);
        return task.getId();
    }

    /**
     * Runs specified task - sends it to first Worker from the queue.
     * 
     * @param taskId id of task to be run
     * @param agent coordinator that will send a message 
     * @throws CodecException 
     * @throws OntologyException
     */
    public void runTask(long taskId, Coordinator agent) throws CodecException, OntologyException
    {
        Task task = (Task) tasks.get(new Long(taskId));

        ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
        msg.addReceiver(task.getNextAgent().getAid());
        msg.setLanguage(codec.getName());
        msg.setOntology(ontology.getName());
        
        Action a = new Action(agent.getAID(), task);
        agent.getContentManager().fillContent(msg, a);

        agent.send(msg);
        Logger.getLogger(Consts.COORDINATOR_LOG).info("Task no. " + task.getId() + " sent to workers , |MRS|="+task.getResultsCount());
    }

    /**
     * Cleans after specified task (after task was performed)
     * @param taskId number of task
     */
    public void cleanAfterTask(long taskId)
    {
        Long id = new Long (taskId);
        Task t = (Task) tasks.get(id);
        tasks.remove(id);
    }
    
    /**
     * Creates a reply message to specified task
     * @param taskId number of task
     * @return build reply-message
     */
    public ACLMessage createReplyMessage(long taskId)
    {
        Task task = (Task)tasks.get(new Long(taskId));
        return task.getMessage().createReply();
    }
}