package com.wawrzek.fumbo.core;

import java.io.*;
import java.util.*;

import org.apache.log4j.*;

import com.hp.hpl.jena.graph.*;
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.rdql.*;
import com.wawrzek.fumbo.*;
import com.wawrzek.fumbo.core.rulepart.*;

/**
 * Class that supports operations on RDF files and model. Wrapper for Jena.
 * 
 * @author Wawrzek
 */
public class RdfEngine
{
    private long url_seq = 1;
    private Model data;
    private Model answer;

    public RdfEngine()
    {
        data = ModelFactory.createDefaultModel();
        answer = ModelFactory.createDefaultModel();
    }

    /**
     * Reads data from given URL to "data"
     * 
     * @param url
     *            url of data source
     */
    public void readDataFromURL(String url)
    {
        Logger.getLogger(Consts.WORKER_LOG).debug("reading from url");
        data.read(url);
    }

    /**
     * Reads data from given string
     * 
     * @param src
     *            data to read
     */
    public void readDataFromString(String src, String baseUri)
    {
        try
        {
            ByteArrayInputStream bis = new ByteArrayInputStream(src.getBytes());
            data.read(bis, baseUri);
            bis.close();
        }
        catch (IOException e)
        {
            e.printStackTrace();
            Logger.getLogger(Consts.COORDINATOR_LOG).error("io error: " + e.getMessage());
        }
    }

    /**
     * submits a RDQL query to the data. Data should be previously read.
     * 
     * @param queryString
     *            RDQL query
     */
    public void query(String queryString)
    {
        Logger.getLogger(Consts.WORKER_LOG).debug("asking query " + queryString);
        Query query = new Query(queryString);
        List vars = query.getBoundVars();
        String var = (String) vars.get(0);

        query.setSource(data);
        QueryExecution qe = new QueryEngine(query);

        QueryResults results = qe.exec();

        if (answer != null)
            answer.close();

        answer = ModelFactory.createDefaultModel();

        ArrayList subjects = new ArrayList();
        while (results.hasNext())
        {
            ResultBinding resb = (ResultBinding) results.next();
            Resource res = (Resource) resb.get(var);
            subjects.add(res);
        }
        results.close();

        int howMany = 0;
        int howManyStatements = 0;
        for (Iterator itr = subjects.iterator(); itr.hasNext();)
        {
            Resource res = (Resource) itr.next();
            howMany++;

            SimpleSelector selector = new SimpleSelector(res, null, (String) null);
            Iterator itr2 = data.listStatements(selector);
            while (itr2.hasNext())
            {
                Statement statement = (Statement) itr2.next();
                answer.add(statement);
                howManyStatements++;
            }
        }
        Logger.getLogger(Consts.WORKER_LOG).debug("" + howManyStatements + " statements added.");
        Logger.getLogger(Consts.WORKER_LOG).debug("" + howMany + " subjects added.");
    }

    /**
     * joins two models - the data and the other, read from the given string
     * 
     * @param other
     */
    public void outerJoin(String other, String baseUri)
    {
        Model mother = ModelFactory.createDefaultModel();
        InputStream is = new ByteArrayInputStream(other.getBytes());
        mother.read(is, baseUri);
        outerJoin(mother);
    }

    /**
     * Joins two models, the data and teh given one.
     * 
     * @param other -
     *            the model to join with data.
     */
    public void outerJoin(Model other)
    {
        answer = answer.add(other);
    }

    /**
     * Returns an RDF as string, serialized answer.
     * 
     * @return RDF data as String
     */
    public String getAnswer()
    {
        OutputStream os;
        os = new ByteArrayOutputStream();
        answer.write(os);
        String result = os.toString();
        try
        {
            os.close();
        }
        catch (IOException e)
        {
            e.printStackTrace();
            Logger.getLogger(Consts.COORDINATOR_LOG).error("io error: " + e.getMessage());
        }

        return result;
    }

    /**
     * Closes datasources, data and answer
     */
    public void close()
    {
        if (data != null)
            data.close();
        if (answer != null)
            answer.close();
    }

    /**
     * Changes the rdql query so that it will have lesser WHERE clause - ignores some part of WHERE clause
     * 
     * @param src source rdql query
     * @param part hich part to be ignored
     * @param whatIgnore literal that will be ignored
     * @return changed rdql query
     */
    public String ignoreConstraints(String src, int part, String whatIgnore)
    {
        Query query = new Query(src);
        List list = query.getTriplePatterns();
        for (Iterator itr = list.iterator(); itr.hasNext();)
        {
            Triple triple = (Triple) itr.next();
            Node node = null;

            if (part == Ignoration.SUBJECT)
                node = triple.getSubject();
            else if (part == Ignoration.PREDICATE)
                node = triple.getPredicate();
            else if (part == Ignoration.OBJECT)
                node = triple.getObject();

            if (node.getURI().equals(whatIgnore))
            {
                itr.remove();
                continue;
            }
        }

        String add = "";
        if (list.size() == 0)
            add = " WHERE (?x ?y ?z)";
        return query.toString() + add;
    }


    /**
     * returns number of subjects 
     * @return number of subjects 
     */
    public int getAnswerCount()
    {
        if (answer == null) {
            return 0;
        }
        int i = 0;
        ResIterator ritr = answer.listSubjects();
        while (ritr.hasNext()){
            Resource r = ritr.nextResource();
            i++;
        }
        return i;
    }
    
    /**
     * returns number of subjects 
     * @return number of subjects 
     */
    public int getDataCount()
    {
        if (data == null) {
            return 0;
        }
        int i = 0;
        ResIterator ritr = data.listSubjects();
        while (ritr.hasNext()){
            Resource r = ritr.nextResource();
            i++;
        }
        return i;
    }
}