package com.wawrzek.fumbo.core;

import java.io.*;

import com.wawrzek.fumbo.core.rulepart.*;
import com.wawrzek.fumbo.tools.*;

/**
 * Class representing the rule, according to which worker works.
 * 
 * @author Wawrzek
 *
 */
public class Rule
{
    /**
     * name of rule-file
     */
    private String fileName;
    /**
     * Condition of a rule ("IF" part)
     */
    private Condition condition;
    /**
     * Command (if exists) of a rule
     */
    private String command;
    /**
     * Meaning (if exists) of rule
     */
    private Meaning meaning;
    
    /**
     * Ignoration (if exists) of rule
     */
    private Ignoration ignoration;

    private final String PREFIX_IF = "IF";
    private final String PREFIX_THEN = "THEN";
    private final String PREFIX_MEANS = "MEANS(";
    private final String PREFIX_RDQL = "RDQL(";
    private final String PREFIX_IGNORE = "IGNORE(";

    public Rule(String fileName) throws IOException, AgentException
    {
        this.fileName = fileName;
        String ruleSource = readFile();
        parseSource(ruleSource);
    }

    /**
     * Reads file to a String
     * @return contet of a file
     * @throws IOException 
     */
    private String readFile() throws IOException
    {
        BufferedReader br = new BufferedReader(new FileReader(fileName));

        String ruleSource = "";
        while (true)
        {
            String temp = br.readLine();
            if (temp == null)
                break;
            ruleSource += temp;
        }
        br.close();
        return ruleSource;
    }

    /**
     * Reads and parses rule from a string
     * @param src text-source of a rule 
     * @throws AgentException
     */
    private void parseSource(String src) throws AgentException
    {
        parseCondition(src);
        parseCommand(src);
    }

    /**
     * Parses string for condition (looking for "IF")
     * @param src String containing Rule
     * @throws AgentException
     */
    private void parseCondition(String src) throws AgentException
    {
        src = src.trim();
        if (!src.startsWith(PREFIX_IF))
            throw new AgentException("error while parsing rule condition");
        int condIndex = PREFIX_IF.length();
        if (condIndex == -1)
            throw new AgentException("error while parsing rule condition");
        int queryIndex = src.indexOf(PREFIX_THEN) + PREFIX_THEN.length();
        if (queryIndex == -1)
            throw new AgentException("error while parsing rule condition");

        String conditionSource = src.substring(condIndex, queryIndex - PREFIX_THEN.length());
        conditionSource = conditionSource.trim();
        condition = ConditionParser.parse(conditionSource);
    }

    /**
     * Parses string for command (after "THEN")
     * @param src String containing Rule
     * @throws AgentException
     */
    private void parseCommand(String src) throws AgentException
    {
        int rdqlIndex = src.indexOf(PREFIX_RDQL);
        if (rdqlIndex != -1)
        {
            parseForRdql(src);
            return;
        }
        int meansIndex = src.indexOf(PREFIX_MEANS);
        if (meansIndex != -1)
        {
            parseForMeans(src);
            return;
        }
        int ignoreIndex = src.indexOf(PREFIX_IGNORE);
        if (ignoreIndex != -1)
        {
            parseForIgnore(src);
            return;
        }
        throw new AgentException("error while parsing rule command");
    }

    /**
     * parses rule for MEANS command
     * @param src string to be parsed
     * @throws AgentException
     */
    private void parseForMeans(String src) throws AgentException
    {
        int firstMeaningIndx = src.indexOf(PREFIX_MEANS) + PREFIX_MEANS.length();
        if (firstMeaningIndx == -1)
            throw new AgentException("error while parsing rule for meaning");
        int sndMeaningIndx = src.indexOf(",");
        if (sndMeaningIndx == -1)
            throw new AgentException("error while parsing rule for meaning");

        String firstMeaning = src.substring(firstMeaningIndx, sndMeaningIndx);
        String secondMeaning = src.substring(sndMeaningIndx + 1, src.indexOf(")"));

        firstMeaning = firstMeaning.trim();
        secondMeaning = secondMeaning.trim();
        meaning = new Meaning(firstMeaning, secondMeaning);

        int meansIndex = src.indexOf(PREFIX_MEANS);
        int bracketIndex = src.indexOf(")");
        command = src.substring(meansIndex, bracketIndex);
    }

    /**
     * parses rule for simple RDQL command
     * @param src string to be parsed
     * @throws AgentException
     */
    private void parseForRdql(String src) throws AgentException
    {
        int rdqlIndex = src.indexOf(PREFIX_RDQL);
        if (rdqlIndex == -1)
            throw new AgentException("error while parsing rule for rdql");
        command = src.substring(rdqlIndex, src.length()).trim();
    }

    /**
     * parses rule for Ignore command
     * @param src string to be parsed
     * @throws AgentException
     */
    private void parseForIgnore(String src) throws AgentException
    {
        int ignoreIndex = src.indexOf(PREFIX_IGNORE);
        if (ignoreIndex == -1)
            throw new AgentException("error while parsing rule for ignore");
        int bracketIndex = src.lastIndexOf(")");
        command = src.substring(ignoreIndex, bracketIndex).trim();
        ignoration = new Ignoration(src.substring(ignoreIndex + PREFIX_IGNORE.length(), bracketIndex));
    }

    /**
     * Returns query that is build upon a rule and orignil query
     * @param originalQuery original query that is base for slightly changed query
     * @return changed query
     */
    public String getQuery(String originalQuery)
    {
        if (command.startsWith(PREFIX_RDQL))
        {
            int rdqlIndex = command.indexOf(PREFIX_RDQL) + PREFIX_RDQL.length();
            int bracketIndex = command.lastIndexOf(")");
            String result = command.substring(rdqlIndex, bracketIndex);
            return result;
        }
        else if (command.startsWith(PREFIX_MEANS))
        {
            String result = meaning.replace(originalQuery);
            return result;
        }
        else if (command.startsWith(PREFIX_IGNORE))
        {
            String result = ignoration.changeQuery(originalQuery);
            return result;
        }
        return null;
    }

    /** 
     * Returns if rule's condition is fulfilled against question and base answer
     * @param question rsql question 
     * @param answer base answer
     * @return true if condition is fulfilled, false otherwise
     * @throws AgentException
     */
    public boolean isTrue(String question, String answer, String baseUri) throws AgentException
    {
        return condition.isTrue(question, answer, baseUri);
    }

    /**
     * Used for tests
     * @param c
     * @param m
     * @param i
     * @return
     */
    public boolean compare(Condition c, Meaning m, Ignoration i)
    {
        boolean cComp = false;
        boolean mComp = false;
        boolean iComp = false;

        if ((condition == null) && (c == null))
            cComp = true;
        if ((condition != null) && (c != null))
            cComp = condition.equals(c);

        if ((meaning == null) && (m == null))
            mComp = true;
        if ((meaning != null) && (m != null))
            mComp = meaning.equals(m);

        if ((ignoration == null) && (i == null))
            iComp = true;
        if ((ignoration != null) && (i != null))
            iComp = ignoration.equals(i);

        return (cComp && mComp && iComp);
    }

    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
        String s = "Rule (";
        s += condition + ", ";
        s += meaning + ", ";
        s += ignoration + ", ";
        if (command.startsWith(PREFIX_RDQL))
            s += command + ", ";
        return s + ")";
    }
}